package checkers.basetype;

import java.util.Properties;

import com.sun.source.tree.CompilationUnitTree;
import com.sun.source.tree.Tree;

import checkers.source.SourceChecker;
import checkers.types.*;

import javax.lang.model.*;
import javax.annotation.processing.*;

/**
 * An abstract {@link SourceChecker} that provides a simple {@link
 * checkers.source.SourceVisitor} implementation for typical assignment and
 * pseudo-assignment checking of annotated types. Most typechecker plug-ins will
 * want to extend this class.
 * 
 * <p>
 * 
 * Subclasses must implement the {@link BaseTypeChecker#isSubtype} method to
 * define the relationships between various annotated and unannotated types in
 * their custom type systems. Note that {@link BaseTypeChecker#isSubtype} must
 * operate on type annotations and not the underlying types (which are checked
 * by the compiler's built-in type checker).
 * 
 * @see BaseTypeChecker#isSubtype(AnnotatedTypeMirror, AnnotatedTypeMirror)
 */
@SupportedSourceVersion(SourceVersion.RELEASE_7)
@SupportedAnnotationTypes({"*"})
public abstract class BaseTypeChecker extends SourceChecker {

    @Override
    protected Properties getMessages() {
        Properties properties = new Properties();
        properties.put("assignability.invalid", 
                "Cannot (re-)assign %s through the reference: %s");
        properties.put("receiver.invalid",
                "incompatible types.\nfound   : %s\nrequired: %s");
        properties.put("type.incompatible", 
                "incompatible types.\nfound   : %s\nrequired: %s");
        properties.put("type.invalid",
                "invalid type, type need to be a supertype of the required " + 
                "type.\nfound   : %s\nrequired: %s");
        properties.put("override.return.invalid", 
                "%s in %s cannot override %s in %s; " + 
                "attempting to use an incompatible return type" + 
                "\nfound   : %s\nrequired: %s");
        properties.put("override.param.invalid", 
                "%s in %s cannot override %s in %s; " +
                "attempting to use an incompatible parameter type" +
                "\nfound   : %s\nrequired: %s");
        properties.put("override.receiver.invalid", 
                "%s in %s cannot override %s in %s; " +
                "attempting to use an incompatible receiver type" +
                "\nfound   : %s\nrequired: %s");
        properties.put("method.invocation.invalid", 
                "call to %s not allowed on the given receiver." + 
                "\nfound   : %s\nrequired: %s");
        return properties;
    }

    @Override
    protected BaseTypeVisitor<?, ?> getSourceVisitor(CompilationUnitTree root) {
        return new BaseTypeVisitor<Void, Void>(this, root);
    }

    /**
     * Tests whether the one annotated type is a subtype of another, with
     * respect to the annotations on these types. Subclasses may wish to ignore
     * annotations that are not related to the type qualifiers they check.
     * 
     * @param sup the parent type
     * @param sub the child type
     * @return true iff {@code sub} is a subtype of {@code sup}
     */
    public abstract boolean isSubtype(AnnotatedTypeMirror sup, AnnotatedTypeMirror sub);

    /**
     * Tests whether the one annotated type is a subtype of another, with
     * respect to the annotations on these types, without descending into
     * generic or array types (i.e., only performing the subtype check on the
     * raw type or outermost array dimension). Subclasses may wish to ignore
     * annotations that are not related to the type qualifiers they check.
     * 
     * @see BaseTypeChecker#isSubtype(AnnotatedTypeMirror, AnnotatedTypeMirror)
     * @param sup the parent type
     * @param sub the child type
     * @return true iff {@code sub} is a subtype of {@code sup}
     * @throws UnsupportedOperationException if this method is not overriden
     */
    public boolean isSubtypeIgnoringTypeParameters(AnnotatedTypeMirror sup, AnnotatedTypeMirror sub) {
        throw new java.lang.UnsupportedOperationException("isSubtypeIgnoringTypeParameters not implemented on this checker");
    }

    /**
     * Tests whether the one annotated type is a valid use of another, with respect
     * to the annotations on these types, without descending into generic or
     * array types (i.e. only performing the validity check on the raw type or 
     * outmost array dimension).
     * 
     * In most cases, the usetype simply needs to be a suptype of elemType.
     * 
     * @param elemType  the type of the class (TypeElement)
     * @param useType   the use of the class (instance type)
     * @return  if the useType is a valid use of elemType
     */
    public boolean isValidUse(AnnotatedTypeMirror elemType, AnnotatedTypeMirror useType) {
        return isSubtype(elemType, useType);
    }

    /**
     * Tests whether the variable accessed is an assignable variable or not, given
     * the current scope
     * 
     * @param varType   the annotated variable type
     * @param variable  the tree access tree
     * @return  true iff variable is assignable in the current scope
     */
    public boolean isAssignable(AnnotatedTypeMirror varType, Tree variable) {
        return true;
    }
}
