package checkers.igj;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.annotation.processing.ProcessingEnvironment;
import javax.lang.model.element.*;
import javax.lang.model.type.*;

import com.sun.source.tree.*;
import com.sun.source.util.SimpleTreeVisitor;

import checkers.igj.quals.I;
import checkers.igj.quals.Immutable;
import checkers.igj.quals.Mutable;
import checkers.igj.quals.ReadOnly;
import checkers.types.*;
import checkers.types.AnnotatedTypeMirror.*;
import checkers.util.AnnotationUtils;
import checkers.util.ElementUtils;
import checkers.util.TreeUtils;
import checkers.util.TypesUtils;

/**
 * Generates {@link AnnotatedTypeMirror} instances accounting for the properties
 * of the IGJ type system. For example, when the
 * {@link IGJAnnotatedTypeFactory#getAnnotatedType(Tree)} method is invoked
 * on a primitive int, it returns the type "@Immutable int".
 */
public class IGJAnnotatedTypeFactory extends AnnotatedTypeFactory{
    /** Adds annotations from tree context before type resolution. */
    private final TreePreAnnotator treePre;

    /** Adds annotations from the resulting type after type resolution. */
    private final TypePostAnnotator typePost;

    /** The various IGJ annotation. */
    protected final AnnotationMirror READONLY, MUTABLE, IMMUTABLE, I,
    PLACE_HOLDER, ASSIGNABLE, ASSIGNS_FIELDS;
    
    protected final AnnotationUtils annoUtils;

    private static final int CACHE_SIZE = 100;

    public IGJAnnotatedTypeFactory(ProcessingEnvironment env,
            CompilationUnitTree root) {
        super(env, root);
        treePre = new TreePreAnnotator();
        typePost = new TypePostAnnotator();
        AnnotationFactory annoFactory = new AnnotationFactory(env);
        READONLY = annoFactory.fromName(ReadOnly.class.getCanonicalName());
        MUTABLE = annoFactory.fromName(Mutable.class.getCanonicalName());
        IMMUTABLE = annoFactory.fromName(Immutable.class.getCanonicalName());
        I = annoFactory.fromName(I.class.getCanonicalName());
        PLACE_HOLDER =
            annoFactory.fromName(IGJPlaceHolder.class.getCanonicalName());
        ASSIGNABLE = annoFactory.fromName("checkers.igj.quals.Assignable");
        ASSIGNS_FIELDS =
            annoFactory.fromName("checkers.igj.quals.AssignsFields");
        annoUtils = new AnnotationUtils(env);
    }

    /**
     * Returns the annotation specifying the immutability type of {@code type}.
     */
    public AnnotationMirror getIGJAnnotation(AnnotatedTypeMirror type) {
        if (type.hasAnnotation(I))
            return type.getAnnotation(I.class.getCanonicalName());
        if (type.hasAnnotation(READONLY))
            return READONLY;
        else if (type.hasAnnotation(MUTABLE))
            return MUTABLE;
        else if (type.hasAnnotation(IMMUTABLE))
            return IMMUTABLE;
        else if (type.hasAnnotation(PLACE_HOLDER))
            return PLACE_HOLDER;
        else if (type.hasAnnotation(ASSIGNS_FIELDS))
            return ASSIGNS_FIELDS;
        else
            return null;
    }
    
    private Map<Tree, AnnotatedTypeMirror> treeCache = createLRUCache(CACHE_SIZE);

    @Override
    public AnnotatedTypeMirror getAnnotatedType(final Tree tree) {
        if (treeCache.containsKey(tree)) {
            return atypes.deepCopy(treeCache.get(tree));
        }
        // Get the underlying expression if the input is parenthesized.
        Tree t = TreeUtils.skipParens(tree);
        
        // Get annotations based on tree properties.
        List<AnnotationMirror> preAnnotations = treePre.visit(t, null);
        
        AnnotatedTypeMirror result = super.getAnnotatedType(t);
        if (preAnnotations != null)
            result.addAnnotations(preAnnotations);
        
        // Try to resolve immutability type variable
        if (result.getKind() != TypeKind.NONE && t instanceof ExpressionTree)
            resolveImmutabilityTypeVar(result, (ExpressionTree) t);
        
        // Add annotations based on type properties.
        typePost.visit(result, tree.getKind() == Tree.Kind.CLASS);
        switch (tree.getKind()) {
        case CLASS:
        case METHOD:
        case VARIABLE:
            treeCache.put(tree, atypes.deepCopy(result));
        }
        return result;
    }
    
    /**
     * Tries to resolve {@code @I} IGJ immutability parameter based on receiver
     * type, and updates the type with the resolved immutability type qualifier.
     * 
     * The method does not change the underlying type of the annotated type, nor
     * does it resolve type variables.
     * 
     * @param type  the type potentially having a immutability type variable
     * @param tree  the expression tree whose type is the provided type
     * @return true if any immutability parameter exists and/or resolved
     */
    public boolean resolveImmutabilityTypeVar(AnnotatedTypeMirror type,
            ExpressionTree tree) {
        // TODO: Clear up this
        if (tree.getKind() == Tree.Kind.IDENTIFIER &&
                ((IdentifierTree)tree).getName().contentEquals("this"))
            return false;
        if (tree.getKind() == Tree.Kind.MEMBER_SELECT &&
                ((MemberSelectTree)tree).getIdentifier().contentEquals("this"))
            return false;
        AnnotatedTypeMirror receiver = getReceiver(tree);
        ImmutabilityTemplateCollector collector =
            new ImmutabilityTemplateCollector();
        
        final Map<String, AnnotationMirror> templateMapping = 
            collector.visit(receiver);

        // There is nothing to resolve
        if (templateMapping == null || templateMapping.isEmpty()) 
            return false;
        
        new ImmutabilityResolver().visit(type, templateMapping);
        
        if (tree instanceof MethodInvocationTree) {
            MethodInvocationTree methodInvoc = (MethodInvocationTree) tree;
            
            Map<String, AnnotationMirror> map = null;
            for (ExpressionTree param : methodInvoc.getArguments()) {
                AnnotatedTypeMirror tm = getAnnotatedType(param);
                map = collector.reduce(map, collector.visit(tm));
            }
            if (map != null && !map.isEmpty())
                new ImmutabilityResolver().visit(type, map);
        }
        
        // This is buggy
        new AnnotatedTypeScanner<Void, Void>() {
            @Override
            public Void visitDeclared(AnnotatedDeclaredType type, Void p) {
                if (type.hasAnnotation(I)) {
                    AnnotationMirror anno = 
                        type.getAnnotation(I.class.getCanonicalName());
                    String s = annoUtils.parseStringValue(anno, 
                            IGJChecker.IMMUTABILITY_KEY);
                    if (!templateMapping.containsKey(s)) {
                        type.removeAnnotation(I);
                        type.addAnnotation(PLACE_HOLDER);
                    }
                }
                return super.visitDeclared(type, p);
            }
        }.visit(type);
        
        return true;
    }

    Map<Element, AnnotatedTypeMirror> elemCache = createLRUCache(CACHE_SIZE);
    
    @Override
    public AnnotatedTypeMirror getAnnotatedType(Element elt) {
        if (elemCache.containsKey(elt))
            return atypes.deepCopy(elemCache.get(elt));
        AnnotatedTypeMirror result = super.getAnnotatedType(elt);
        // Add annotations based on type properties.
        typePost.visit(result, elt.getKind().isClass() || elt.getKind().isInterface());
        elemCache.put(elt, atypes.deepCopy(result));
        return result;
    }

    private class TreePreAnnotator
    extends SimpleTreeVisitor<List<AnnotationMirror>, Void> {
        @Override
        public List<AnnotationMirror> visitNewClass(NewClassTree node,
        		Void p) {
        	// If no immutability is specified, use a place holder
            AnnotatedDeclaredType type = (AnnotatedDeclaredType)fromExpression(node);
            AnnotatedTypeMirror ct = fromElement(type.getUnderlyingType().asElement());
            
            if (!hasImmutabilityType(type) && !hasImmutabilityType(ct))
                return Collections.singletonList(PLACE_HOLDER);

            return null;
        }
    }

    /**
     * A class for adding annotations to a type after initial type resolution.
     * Used here for resolving this-mutability.
     */
    private class TypePostAnnotator  extends AnnotatedTypeScanner<Void, Boolean> {
    	
        /** All primitives are immutable **/
        @Override
        public Void visitPrimitive(AnnotatedPrimitiveType type, Boolean p) {
        	if (!hasImmutabilityType(type))
        		type.addAnnotation(PLACE_HOLDER);
            return super.visitPrimitive(type, p);
        }

        /** Returns {@code true} if type is a boxed primitive type **/
        protected boolean isBoxedImmutable(AnnotatedDeclaredType type) {
            DeclaredType underlyingType = type.getUnderlyingType();
            Name qualifiedName = TypesUtils.getQualifiedName(underlyingType);

            return (qualifiedName.contentEquals("java.lang.String") ||
                    qualifiedName.contentEquals("java.lang.Boolean") ||
                    qualifiedName.contentEquals("java.lang.Byte") ||
                    qualifiedName.contentEquals("java.lang.Character") ||
                    qualifiedName.contentEquals("java.lang.Enum") ||
                    qualifiedName.contentEquals("java.lang.Short") ||
                    qualifiedName.contentEquals("java.lang.Integer") ||
                    qualifiedName.contentEquals("java.lang.Number") ||
                    qualifiedName.contentEquals("java.lang.Long") ||
                    qualifiedName.contentEquals("java.lang.Double") ||
                    qualifiedName.contentEquals("java.lang.Float") ||
                    qualifiedName.contentEquals("java.lang.Class") ||
                    qualifiedName.contentEquals("java.lang.Void"));
        }

        @Override
        public Void visitDeclared(AnnotatedDeclaredType type, Boolean p) {
        	if (!hasImmutabilityType(type)) {
        		// By default:
        		//  - Classes: 				Mutable
        		//  - Interfaces: 			Place holder
        		//  - Enum & Annotations:	Immutable
        		
        	    Element elem = type.getUnderlyingType().asElement();
        		AnnotatedDeclaredType elemType = (AnnotatedDeclaredType)
        		    fromElement(elem);

        		if (isBoxedImmutable(type))
        			type.addAnnotation(PLACE_HOLDER);
        		else if (elem.getSimpleName().contentEquals("Object"))
        		    type.addAnnotation(PLACE_HOLDER);
        		else if (hasImmutabilityType(elemType)) {
        		    if (elemType.hasAnnotation(IMMUTABLE))
        		        type.addAnnotation(IMMUTABLE);
        		    else
        		        type.addAnnotation(MUTABLE);
        		} else if (elem.getKind() == ElementKind.CLASS)
        		    type.addAnnotation(MUTABLE);
        		else if (elem.getKind() == ElementKind.INTERFACE) {
        		    type.addAnnotation(p ? PLACE_HOLDER : MUTABLE);
        		} else if (elem.getKind() == ElementKind.ENUM ||
        				elem.getKind() == ElementKind.ANNOTATION_TYPE)
        			type.addAnnotation(IMMUTABLE);
        		else
        			assert false : "Shouldn't be here";
        	}
        	
        	return super.visitDeclared(type, p);
        }
        
        @Override
        public Void visitNull(AnnotatedNullType type, Boolean p) {
            type.addAnnotation(PLACE_HOLDER);
        	return super.visitNull(type, p);
        }
        
        @Override
        public Void visitExecutable(AnnotatedExecutableType type, Boolean p) {
        	ExecutableElement elem = (ExecutableElement) type.getElement();

        	// Check Receiver
        	if (!hasImmutabilityType(type.getReceiverType())) {
        		AnnotatedDeclaredType receiverType = type.getReceiverType();
        		TypeElement classImmutability = 
        			(TypeElement) elem.getEnclosingElement();
        		AnnotatedDeclaredType classType = 
        			(AnnotatedDeclaredType) getAnnotatedType(classImmutability);

        		if (ElementUtils.isObject(classImmutability) ||
        		        TypesUtils.isObject(receiverType.getUnderlyingType()))
        		    receiverType.addAnnotation(PLACE_HOLDER);
        		else if (elem.getKind() == ElementKind.CONSTRUCTOR) {
        		    if (classType.hasAnnotation(MUTABLE))
        		        receiverType.addAnnotation(MUTABLE);
        		    else
        		        receiverType.addAnnotation(ASSIGNS_FIELDS);
        		} else if (classType.hasAnnotation(IMMUTABLE)) {
        		    receiverType.addAnnotation(READONLY);
        		} else
        			receiverType.addAnnotation(MUTABLE);
        	}
        	
        	return super.visitExecutable(type, p);
        }
        
        @Override
        public Void visitArray(AnnotatedArrayType type, Boolean p) {
            if (!hasImmutabilityType(type))
                type.addAnnotation(MUTABLE);
            return super.visitArray(type, p);
        }
        
        @Override
        public Void visitTypeVariable(AnnotatedTypeVariable type, Boolean p) {
            if (!hasImmutabilityType(type)) {
                if (type.getUpperBound() != null && hasImmutabilityType(type.getUpperBound()))
                    type.addAnnotation(getIGJAnnotation(type.getUpperBound()));
                else
                    type.addAnnotation(PLACE_HOLDER);
            }
            return super.visitTypeVariable(type, p);
        }
        
        @Override
        public Void visitWildcard(AnnotatedWildcardType type, Boolean p) {
            if (!hasImmutabilityType(type)) {
                if (type.getExtendsBound() != null && hasImmutabilityType(type.getExtendsBound()))
                    type.addAnnotation(getIGJAnnotation(type.getExtendsBound()));
                else
                    type.addAnnotation(PLACE_HOLDER);
            }
            return super.visitWildcard(type, p);
        }
    }

    /**
     * @param type  an annotated type mirror
     * @return  true iff the type is specified an immutability type, 
     *          false otherwise
     */
    public boolean hasImmutabilityType(AnnotatedTypeMirror type) {
        return getIGJAnnotation(type) != null;
    }
    
    /**
     * A Helper class that tries to resolve the immutability type variable,
     * as the type variable is assigned to the most restricted immutability
     */
    private class ImmutabilityTemplateCollector
    extends AnnotatedTypeScanner<Map<String, AnnotationMirror>, Void> {
        
        @Override
        public Map<String, AnnotationMirror> reduce(Map<String, AnnotationMirror> r1,
                Map<String, AnnotationMirror> r2) {
            Map<String, AnnotationMirror> result = 
                new HashMap<String, AnnotationMirror>();
            
            if (r1 != null)
                result.putAll(r1);
            
            if (r2 != null) {
                // Need to be careful about overlap
                for (String key : r2.keySet()) {
                    if (!result.containsKey(key))
                        result.put(key, r2.get(key));
                    else if (!result.get(key).equals(r2.get(key)))
                        result.put(key, READONLY);
                }
            }
            return result;
        }
        
        @Override
        public Map<String, AnnotationMirror> visitDeclared(
                AnnotatedDeclaredType type, Void p) {
            Map<String, AnnotationMirror> result =
                new HashMap<String, AnnotationMirror>();
            
            TypeElement typeElement = 
                (TypeElement) type.getUnderlyingType().asElement();
            AnnotatedDeclaredType dcType = 
                (AnnotatedDeclaredType) fromElement(typeElement);
            
            if (dcType.hasAnnotation(I)) {
                String immutableString = 
                    annoUtils.parseStringValue(getIGJAnnotation(dcType), 
                            IGJChecker.IMMUTABILITY_KEY);
                AnnotationMirror immutability = getIGJAnnotation(type);
                // Assertion failes some times
//                assert immutability != null;
                if ((immutability != null) && (!type.hasAnnotation(ASSIGNS_FIELDS)))
                    result.put(immutableString, immutability);
            }

            return reduce(result, super.visitDeclared(type, p));
        }
    }
    
    /**
     * A helper class that resolves the immutability on a types based on a
     * provided mapping.
     * 
     * It returns a set of the annotations that were inserted. This is important
     * to recognize which immutability type variables were resolved and which
     * are to be made into place holder.
     */
    private class ImmutabilityResolver extends
    AnnotatedTypeScanner<Set<AnnotationMirror>, Map<String, AnnotationMirror>> {
        
        @Override
        public Set<AnnotationMirror> reduce(Set<AnnotationMirror> r1,
                Set<AnnotationMirror> r2) {
            Set<AnnotationMirror> result = new HashSet<AnnotationMirror>();
            if (r1 != null)
                result.addAll(r1);
            if (r2 != null)
                result.addAll(r2);
            return result;
        }
        
        @Override
        public Set<AnnotationMirror> visitDeclared(AnnotatedDeclaredType type, 
                Map<String, AnnotationMirror> p) {
            Set<AnnotationMirror> result = new HashSet<AnnotationMirror>();

            if (type.hasAnnotation(I)) {
                String immutableString = 
                    annoUtils.parseStringValue(getIGJAnnotation(type), 
                            IGJChecker.IMMUTABILITY_KEY);
                type.removeAnnotation(I);
                if (p.containsKey(immutableString)) {
                    type.addAnnotation(p.get(immutableString));
                    result.add(p.get(immutableString));
                }
            }
            
            return reduce(result, super.visitDeclared(type, p));
        }
    }

    @Override
    public void postDirectSuperTypes(AnnotatedTypeMirror type,
            List<? extends AnnotatedTypeMirror> supertypes) {

        // Try to resolve @I
        ImmutabilityTemplateCollector collector =
            new ImmutabilityTemplateCollector();
        
        Map<String, AnnotationMirror> templateMapping = 
            collector.visit(type);
        ImmutabilityResolver resolver = new ImmutabilityResolver();
        
        for (AnnotatedTypeMirror t : supertypes) {
            if (templateMapping != null && !templateMapping.isEmpty())
                resolver.visit(t, templateMapping);
            this.typePost.visit(t, false);
        }
    }

    @Override
    protected void annotateInheritedFromClass(@Mutable AnnotatedTypeMirror type) {
        // Do nothing
    }

    /**
     * Returns a type of the class and the field {@code this}.
     */
    @Override
    public AnnotatedDeclaredType getSelfType(Tree tree) {
        AnnotatedDeclaredType act = visitorState.getClassType();
        AnnotatedDeclaredType methodReceiver = visitorState.getMethodReceiver();
        
        if (methodReceiver == null)
            return act;
        // Are we in a mutable or Immutable scope
        if (methodReceiver.hasAnnotation(MUTABLE) ||
                methodReceiver.hasAnnotation(IMMUTABLE)) {
            return methodReceiver;
        } else if (act.hasAnnotation(I)) {
            if (methodReceiver.hasAnnotation(ASSIGNS_FIELDS))
                act.addAnnotation(ASSIGNS_FIELDS);
            return act;
        } else if (act.hasAnnotation(IMMUTABLE)) {
            return act;
        } else // TODO: Figure out when
            return methodReceiver;
    }

    @Override
    protected Collection<AnnotationMirror> unify(Collection<AnnotationMirror> c1,
            Collection<AnnotationMirror> c2) {
        Map<String, AnnotationMirror> first = 
            new HashMap<String, AnnotationMirror>();
        for (AnnotationMirror anno : c1)
            first.put(AnnotationUtils.annotationName(anno), anno);
        Set<String> second = new HashSet<String>();
        for (AnnotationMirror anno : c2)
            second.add(AnnotationUtils.annotationName(anno));
        first.keySet().retainAll(second);
        // We don't have any igj type add ReadOnly
        if (!first.containsKey(I.class.getCanonicalName()) &&
                !first.containsKey(Mutable.class.getCanonicalName()) &&
                !first.containsKey(Immutable.class.getCanonicalName()) &&
                !first.containsKey(IGJPlaceHolder.class.getCanonicalName()))
            first.put(ReadOnly.class.getCanonicalName(), READONLY);
        return first.values();
    }

}
