package checkers.interned;

import checkers.basetype.*;
import checkers.source.*;
import checkers.types.*;
import checkers.util.SimpleSubtypeRelation;

import java.util.*;

import javax.annotation.processing.*;
import javax.lang.model.*;
import javax.lang.model.element.*;

import com.sun.source.tree.*;

/**
 * A typechecker plug-in for the {@link checkers.quals.Interned} qualifier that
 * finds (and verifies the absence of) equality-testing and interning errors.
 * 
 * <p>
 * 
 * The {@link checkers.quals.Interned} annotation indicates that a variable
 * refers to the canonical instance of an object, meaning that it is safe to
 * compare that object using the "==" operator. This plugin suggests using "=="
 * instead of ".equals" where possible, and warns whenever "==" is used in cases
 * where one or both operands are not {@link checkers.quals.Interned}.
 * 
 * @manual #interned Interned checker
 */
@SupportedAnnotationTypes({"*"})
@SupportedSourceVersion(SourceVersion.RELEASE_7)
@SupportedLintOptions({"dotequals"})
// TODO: move some of these options to annotations
public final class InternedChecker extends BaseTypeChecker {

    /** For generating annotations. */
    private AnnotationFactory annoFactory;
    
    /** The subtype relation behind the {@link #isSubtype} method. */
    private SimpleSubtypeRelation relation;
    
    /** Represents the {@link Interned} annotation. */
    private AnnotationMirror INTERNED;

    @Override
    public synchronized void init(ProcessingEnvironment processingEnv) {
        super.init(processingEnv);
        this.annoFactory = new AnnotationFactory(this.env);
        this.INTERNED = annoFactory.fromName("checkers.quals.Interned");
        this.relation = new SimpleSubtypeRelation(INTERNED, null);
    }

    @Override
    public boolean isSubtype(AnnotatedTypeMirror lhs, AnnotatedTypeMirror rhs) {
        return relation.isSubtype(lhs, rhs);
    }

    @Override
    protected String getDefaultSkipPattern() {
        return "";
    }

    @Override
    protected Properties getMessages() {
        Properties msgs = new Properties(super.getMessages());
        msgs.setProperty("not.interned",
                "comparison requires two @Interned types");
        msgs.setProperty("unnecessary.equals",
                "use of .equals can be safely replaced by ==/!=");
        return msgs;
    }

    @Override
    public BaseTypeVisitor<?, ?> getSourceVisitor(CompilationUnitTree root) {
        return new InternedVisitor(this, root);
    }

    @Override
    public AnnotatedTypeFactory getFactory(ProcessingEnvironment env, CompilationUnitTree root) {
        return new InternedAnnotatedTypeFactory(env, root);
    }

    @Override
    protected String getSuppressWarningsKey() {
        return "interned";
    }
}

