package checkers.javari;

import java.util.*;

import javax.annotation.processing.ProcessingEnvironment;
import javax.lang.model.element.*;

import checkers.types.*;
import checkers.util.TreeUtils;
import static checkers.types.AnnotatedTypeMirror.*;

import com.sun.source.tree.*;
import com.sun.source.util.SimpleTreeVisitor;
import javax.lang.model.util.SimpleElementVisitor6;
import com.sun.tools.javac.tree.JCTree;

import checkers.quals.*;

/**
 * Generates AnnotatedTypeMirrors with Javari annotations from Trees and Elements.
 */
public class JavariAnnotatedTypeFactory extends AnnotatedTypeFactory {

    /** Adds annotations from tree context before type resolution. */
    private final TreePreAnnotator treePre;
    
    /** Adds annotations from the resulting type after type resolution. */
    private final TypePostAnnotator typePost;

    /** The Javari annotations. */
    private final AnnotationMirror READONLY, MUTABLE, ROMAYBE, QREADONLY, ASSIGNABLE;

    /**
     * Creates a new {@link InternedAnnotatedTypeFactory} that operates on a
     * particular AST.
     *
     * @param env the current processing environment for this annotation
     *                processor
     * @param root the AST on which this type factory operates
     */
    public JavariAnnotatedTypeFactory(ProcessingEnvironment env,
        CompilationUnitTree root) {
        super(env, root);
        this.treePre = new TreePreAnnotator();
        this.typePost = new TypePostAnnotator();
        this.READONLY = annotations.fromName("checkers.quals.ReadOnly");
        this.MUTABLE = annotations.fromName("checkers.quals.Mutable");
        this.ROMAYBE = annotations.fromName("checkers.quals.RoMaybe");
        this.QREADONLY = annotations.fromName("checkers.quals.QReadOnly");
        this.ASSIGNABLE = annotations.fromName("checkers.quals.Assignable");
    }

    @Override
    public AnnotatedTypeMirror getAnnotatedType(Tree tree) {
    	tree = TreeUtils.skipParens(tree);
        AnnotatedTypeMirror result = super.getAnnotatedType(tree);
        annotateInheritedFromClass(result);
        return result;
    }

    @Override
    public AnnotatedTypeMirror getAnnotatedType(Element elt) {
        AnnotatedTypeMirror result = super.getAnnotatedType(elt);
        annotateInheritedFromClass(result);
        return result;
    }

    @Override
    protected void annotateImplicit(Tree tree, /*@Mutable*/ AnnotatedTypeMirror type) {
        treePre.visit(tree, type);
    }

    @Override
    protected void annotateInheritedFromClass(/*@Mutable*/ AnnotatedTypeMirror type) {
        typePost.visit(type);
    }

    /**
     * A visitor to get the annotations from a tree.
     */
    private class TreePreAnnotator extends SimpleTreeVisitor<Void, AnnotatedTypeMirror> {

        /** Default action: do nothing. */
        @Override
        public Void defaultAction(Tree node, AnnotatedTypeMirror p) {
            return null;
        }


        /** Ensures the literal has a {code @Mutable} annotation. */
        @Override
        public Void visitLiteral(LiteralTree node, AnnotatedTypeMirror p) {
            p.addAnnotation(MUTABLE);
            return super.visitLiteral(node, p);
        }


        /** Ensures the result has a {code @Mutable} annotation. */
        @Override
        public Void visitBinary(BinaryTree node, AnnotatedTypeMirror p) {
            p.addAnnotation(MUTABLE);
            return super.visitBinary(node, p);
        }


        /** Ensures the result has a {code @Mutable} annotation. */
        @Override
        public Void visitUnary(UnaryTree node, AnnotatedTypeMirror p) {
            p.addAnnotation(MUTABLE);
            return super.visitUnary(node, p);
        }


        /**
         * Makes a copy of the list of annotations of the true branch,
         * then adds to that new list all annotations from the false branch
         * if not present yet.
         */
        @Override
        public Void visitConditionalExpression(ConditionalExpressionTree node, AnnotatedTypeMirror p) {
            AnnotatedTypeMirror trueType = getAnnotatedType(node.getTrueExpression());
            if (trueType.getKind().isPrimitive()) return super.visitConditionalExpression(node, p);

            AnnotatedTypeMirror falseType = getAnnotatedType(node.getFalseExpression());

            if (trueType.hasAnnotation(QREADONLY) || falseType.hasAnnotation(QREADONLY))
                p.addAnnotation(QREADONLY);
            else if (trueType.hasAnnotation(READONLY) || falseType.hasAnnotation(READONLY))
                p.addAnnotation(READONLY);
            else if (trueType.hasAnnotation(ROMAYBE) || falseType.hasAnnotation(ROMAYBE))
                p.addAnnotation(ROMAYBE);
            else
                p.addAnnotation(MUTABLE);

            return super.visitConditionalExpression(node, p);
        }

        /**
         * Resolves the return type of a method invocation, if it is annotated with {code @RoMaybe}, if possible.
         */
        @Override
        public Void visitMethodInvocation(MethodInvocationTree node, AnnotatedTypeMirror p) {

            AnnotatedExecutableType executableType =
                (AnnotatedExecutableType) getAnnotatedType(node.getMethodSelect());
            ExecutableElement executableElt = (ExecutableElement) executableType.getElement();
            AnnotatedTypeMirror returnType = executableType.getReturnType();


            if (returnType.hasAnnotation(ROMAYBE)) {
                List<AnnotatedTypeMirror> parameterTypes = executableType.getParameterTypes();
                List<? extends ExpressionTree> argumentTrees = node.getArguments();
                AnnotatedTypeMirror receiverType = executableType.getReceiverType();

                boolean allMutable = true, allRoMaybe = true;

                // look at parameters and arguments
                // TODO: deal with varargs parameters
                for (int i = 0; i < parameterTypes.size(); i++) {
                    AnnotatedTypeMirror pType = parameterTypes.get(i);
                    // only look at it if parameter is RoMaybe
                    if (pType.hasAnnotation(ROMAYBE)) {
                        AnnotatedTypeMirror aType = getAnnotatedType(argumentTrees.get(i));

                        // if argument has no Mutable and has annotations, not all are Mutable
                        if (!aType.hasAnnotation(MUTABLE) && !aType.getAnnotations().isEmpty())
                            allMutable = false;

                        // if argument has no RoMaybe, not all args are RoMaybe
                        if (!aType.hasAnnotation(ROMAYBE))
                            allRoMaybe = false;

                    }
                }

                // look at receiver type and reference
                if (receiverType.hasAnnotation(ROMAYBE)) {
                    // if MemberSelectTree, we can just look at the expression tree
                    ExpressionTree exprTree = node.getMethodSelect();
                    if (exprTree.getKind() == Tree.Kind.MEMBER_SELECT) {
                        MemberSelectTree mst = (MemberSelectTree) exprTree;
                        AnnotatedTypeMirror referenceType = getAnnotatedType(mst.getExpression());

                        if (referenceType.hasAnnotation(READONLY)) {
                            allMutable = false;
                            allRoMaybe = false;
                        } else if (referenceType.hasAnnotation(ROMAYBE)) {
                            allMutable = false;
                        }

                    } else {
                        // not MemberSelectTree, get context from method's enclosing class
                        AnnotatedTypeMirror classType
                            = getAnnotatedType(executableElt.getEnclosingElement());
                        if (classType.hasAnnotation(READONLY)) {
                            allMutable = false;
                            allRoMaybe = false;
                        } else if (classType.hasAnnotation(ROMAYBE)) {
                            allMutable = false;
                        }
                    }
                }

                // now, add appropriate type to return list
                if (allMutable) p.addAnnotation(MUTABLE);
                else if (allRoMaybe) p.addAnnotation(ROMAYBE);
                else p.addAnnotation(READONLY);

            } else { // return type has no RoMaybe, just get the annotated return type
                if (returnType.hasAnnotation(READONLY)) p.addAnnotation(READONLY);
                else if (!returnType.getKind().isPrimitive()) p.addAnnotation(MUTABLE);
            }
            return super.visitMethodInvocation(node, p);
        }


        /** Returns list of annotations on the new class constructor; resolves {code @RoMaybe} if applicable. */
        @Override
        public Void visitNewClass(NewClassTree node, AnnotatedTypeMirror p) {

            Element constructorElt = ((JCTree.JCNewClass)node).constructor;
            AnnotatedExecutableType executableType =
                (AnnotatedExecutableType) getAnnotatedType(constructorElt);
            AnnotatedTypeMirror receiverType = executableType.getReceiverType();

            // if receiver type has RoMaybe, we must resolve it by looking at the arguments
            if (receiverType.hasAnnotation(ROMAYBE)) {
                List<AnnotatedTypeMirror> parameterTypes = executableType.getParameterTypes();
                List<? extends ExpressionTree> argumentTrees = node.getArguments();

                boolean allMutable = true, allRoMaybe = true;

                for (int i = 0; i < parameterTypes.size(); i++) {
                    AnnotatedTypeMirror pType = parameterTypes.get(i);
                    if (pType.hasAnnotation(ROMAYBE)) {
                        AnnotatedTypeMirror aType = getAnnotatedType(argumentTrees.get(i));

                        if (aType.hasAnnotation(READONLY) || aType.hasAnnotation(QREADONLY))
                            allMutable = false;

                        if (!(aType.hasAnnotation(ROMAYBE)
                              && !aType.hasAnnotation(READONLY)
                              && !aType.hasAnnotation(MUTABLE)
                              && !aType.hasAnnotation(QREADONLY)))
                            allRoMaybe = false;
                    }
                }

                if (allMutable) {
                    if (!p.hasAnnotation(MUTABLE)) p.addAnnotation(MUTABLE);
                } else if (allRoMaybe) {
                    if (!p.hasAnnotation(ROMAYBE)) p.addAnnotation(ROMAYBE);
                } else {
                    if (!p.hasAnnotation(READONLY)) p.addAnnotation(READONLY);
                }

            } else {   // if receiver has no romaybe, just return the receiver type
                if (receiverType.hasAnnotation(READONLY)) p.addAnnotation(READONLY);
                if (receiverType.hasAnnotation(MUTABLE)) p.addAnnotation(MUTABLE);
            }
            
            return super.visitNewClass(node, p);
        }

        /**
         * Selects the appropriate annotation from the inner type, or inherits
         * it from the outer type according to the Javari typesystem.
         */
        @Override
        public Void visitMemberSelect(MemberSelectTree node, AnnotatedTypeMirror p) {

        	AnnotatedTypeMirror outerType = getAnnotatedType(node.getExpression());
            AnnotatedTypeMirror innerType = getAnnotatedType(elementFromUse(node));

            if (innerType.hasAnnotation(READONLY))
                p.addAnnotation(READONLY);
            else if (innerType.hasAnnotation(MUTABLE))
                p.addAnnotation(MUTABLE);
            else if (innerType.hasAnnotation(ROMAYBE) && outerType.hasAnnotation(ROMAYBE))
                p.addAnnotation(ROMAYBE);
            else if (outerType.hasAnnotation(READONLY))
                p.addAnnotation(READONLY);
            else if (outerType.hasAnnotation(MUTABLE))
                p.addAnnotation(MUTABLE);
            else if (outerType.hasAnnotation(ROMAYBE))
                p.addAnnotation(ROMAYBE);

            return super.visitMemberSelect(node, p);
        }

    }


    /**
     * Responsible for adding implicit (not written on the code) annotations.
     */
    private class TypePostAnnotator extends AnnotatedTypeScanner<Void, Void> {

        /**
         * Passes down the enclosing class annotation, if any; else try
         * to retrieve from annotation if available, from enclosing
         * class if field, from supertype if class or interface,
         * or mutable otherwise.
         */
        @Override
        public Void scan(AnnotatedTypeMirror type, Void p) {
            // we don't want to add annotations if there are annotations already
            if (type.getAnnotations().isEmpty() && type.getElement() != null) {
                ElementKind eltKind = type.getElement().getKind();

                if (eltKind == ElementKind.LOCAL_VARIABLE) {
                    type.addAnnotation(MUTABLE);

                } else if (eltKind.isField()) {
                    // inherit annotations from enclosing class
                    AnnotatedTypeMirror enclosingType =
                        getAnnotatedType(type.getElement().getEnclosingElement());

                    if (enclosingType.hasAnnotation(READONLY))
                        type.addAnnotation(READONLY);

                } else if (eltKind.isClass()) {
                    AnnotatedTypeMirror superType
                        = ((AnnotatedDeclaredType) type).getSuperclass();

                    // if Object, supertype has underlying type NoType with
                    // TypeKind NONE, and no annotations

                    if (superType.hasAnnotation(MUTABLE))
                        type.addAnnotation(MUTABLE);
                    else if (superType.hasAnnotation(READONLY))
                        type.addAnnotation(READONLY);
                    else if (superType.hasAnnotation(ROMAYBE))
                        type.addAnnotation(ROMAYBE);
                    else
                        type.addAnnotation(MUTABLE);

                } else if (eltKind.isInterface()) {

                    AnnotatedDeclaredType iType = (AnnotatedDeclaredType) type;
                    boolean annotated = false;
                    for (AnnotatedTypeMirror i : iType.getInterfaces()) {
                        if (i.hasAnnotation(READONLY)) {
                            type.addAnnotation(READONLY);
                            break;
                        }
                    }
                    if (!annotated) {
                        for (AnnotatedTypeMirror i : iType.getInterfaces()) {
                            if (i.hasAnnotation(ROMAYBE)) {
                                type.addAnnotation(ROMAYBE);
                                annotated = true;
                                break;
                            }
                        }
                    }

                }

            }
            return super.scan(type, p);
        }
    }
}
