package checkers.javari;

import java.util.*;

import javax.lang.model.element.*;
import javax.lang.model.type.*;
import javax.lang.model.util.ElementFilter;

import checkers.source.*;
import checkers.basetype.*;
import checkers.types.*;
import checkers.types.AnnotatedTypeMirror.*;

import com.sun.source.tree.*;
import com.sun.tools.javac.tree.JCTree;

/**
 * A type-checking visitor for the Javari mutability annotations
 * ({@code @ReadOnly}, {@code @Mutable} and {@code @Assignable}) that
 * extends BaseTypeVisitor.
 *
 * @see BaseTypeVisitor
 */
public class JavariVisitor extends BaseTypeVisitor<Void, Void> {

    final private AnnotationMirror MUTABLE, READONLY,
        ROMAYBE, QREADONLY, ASSIGNABLE;

    /**
     * Creates a new visitor for type-checking the Javari mutability
     * annotations.
     *
     * @param checker the {@link JavariChecker} to use
     * @param root the root of the input program's AST to check
     */
    public JavariVisitor(JavariChecker checker, CompilationUnitTree root) {
        super(checker, root);
        this.checker = checker;
        READONLY = checker.READONLY;
        MUTABLE = checker.MUTABLE;
        ROMAYBE = checker.ROMAYBE;
        QREADONLY = checker.QREADONLY;
        ASSIGNABLE = checker.ASSIGNABLE;
    }

    /**
     * Modifies p to the appropriate values, according to super
     * classes, interfaces, and outer classes.
     *
     * @see com.sun.source.util.TreeScanner#visitClass(com.sun.source.tree.ClassTree,java.lang.Object)
     */
    @Override
    public Void visitClass(ClassTree node, Void p) {

        boolean readonly = false, mutable = false, romaybe = false;
        AnnotatedTypeMirror type = factory.getAnnotatedType(node);
        Tree ext = node.getExtendsClause();

        if (ext != null){
            AnnotatedTypeMirror extendsType = factory.getAnnotatedType(ext);

            if (!checker.isSubtype(extendsType, type))
                checker.report(Result.failure("extends.invalid"), node.getExtendsClause());

            if (extendsType.hasAnnotation(READONLY))
                readonly = true;
            if (extendsType.hasAnnotation(MUTABLE))
                mutable = true;
            if (extendsType.hasAnnotation(ROMAYBE))
                romaybe = true;
        }

        List<? extends Tree> imps = node.getImplementsClause();
        for(Tree it : imps) {
            AnnotatedTypeMirror iType = factory.getAnnotatedType(it);
            if (!checker.isSubtype(iType, type))
                checker.report(Result.failure("implements.invalid"), it);

            if (iType.hasAnnotation(READONLY))
                readonly = true;
            if (iType.hasAnnotation(MUTABLE))
                mutable = true;
            if (iType.hasAnnotation(ROMAYBE))
                romaybe = true;
        }

        if (type.hasAnnotation(READONLY))
            readonly = true;
        else if (type.hasAnnotation(MUTABLE))
            mutable = true;
        else if (type.hasAnnotation(ROMAYBE))
            romaybe = true;

        if (!(readonly || mutable || romaybe))
            mutable = true;

        if (romaybe) {
            if (readonly || mutable)
                checker.report(Result.failure("romaybe.only"), node);
            if (visitorState.getMethodReceiver() == null || !visitorState.getMethodReceiver().hasAnnotation(ROMAYBE))
                checker.report(Result.failure("romaybe.type"), node);

        } else if (readonly && mutable) {
            checker.report(Result.failure("ro.and.mutable"), node);
        }

        return super.visitClass(node, p);
    }

    /**
     * Allows a variable marked as ReadOnly and RoMaybe to be returned on a RoMaybe method.
     *
     * @see com.sun.source.util.TreeScanner#visitReturn(con.sun.source.tree.ReturnTree,java.lang.Object)
     */
    @Override
    public Void visitReturn(ReturnTree node, Void p) {
        // Don't try to check return expressions for void methods.
        if (node.getExpression() == null)
            return super.visitReturn(node, p);

        AnnotatedTypeMirror returnValue = factory.getAnnotatedType(node.getExpression());

        if (returnValue.hasAnnotation(ROMAYBE))
            return null;

        return super.visitReturn(node, p);
    }

    /**
     * Modifies the parameter p to its appropriate value. Checks:
     *
     * <ul>
     *  <li> the return type cannot be primitive and marked with ReadOnly
     *  <li> whether the method overrides a method illegally
     *  <li> whether the receiver type is invalid (Mutable inside ReadOnly)
     * </ul>
     *
     * @see com.sun.source.util.TreeScanner#visitMethod(com.sun.source.tree.MethodTree,java.lang.Object)
     */
    @Override
    public Void visitMethod(MethodTree node, Void p) {

        AnnotatedExecutableType method = (AnnotatedExecutableType) factory.getAnnotatedType(node);
        AnnotatedTypeMirror receiverType = method.getReceiverType();

        // check if receiver is invalid
        if ((receiverType.hasAnnotation(MUTABLE)
             || receiverType.hasAnnotation(ROMAYBE))
            && visitorState.getClassType().hasAnnotation(READONLY))
            checker.report(Result.failure("receiver.invalid"), node);

        ExecutableElement methodElt = (ExecutableElement) method.getElement();
        ExecutableElement oMethodElt = null;
        javax.annotation.processing.ProcessingEnvironment env = checker.getProcessingEnvironment();

        TypeElement classElt = (TypeElement) methodElt.getEnclosingElement();
        TypeMirror classType = classElt.asType();
        List<? extends TypeMirror> supers
            = env.getTypeUtils().directSupertypes(classType);

        // climb up the tree until you find the method, or until there is nowhere to climb to
        while (oMethodElt == null) {
            for(TypeMirror t : supers) {
                TypeElement superElt
                    = (TypeElement) ((DeclaredType)t).asElement();
                List<ExecutableElement> methods
                    = ElementFilter.methodsIn(superElt.getEnclosedElements());
                javax.lang.model.util.Elements eltUtils = env.getElementUtils();
                for (ExecutableElement e : methods) {
                    if (eltUtils.overrides(methodElt, e, classElt)) {
                        oMethodElt = e;
                        break;
                    }
                }
            }
            if (oMethodElt == null) {
                if (supers.isEmpty()) break;
                supers = env.getTypeUtils().directSupertypes(supers.get(0));
            }
        }


        if (oMethodElt != null) {

            AnnotatedDeclaredType enclosingType
                = (AnnotatedDeclaredType) factory.getAnnotatedType(classElt);
            AnnotatedExecutableType overridenMethod
                = (AnnotatedExecutableType) factory.getAnnotatedType(oMethodElt);
            AnnotatedDeclaredType overridenType
                = (AnnotatedDeclaredType) factory.getAnnotatedType(oMethodElt.getEnclosingElement());

            checkOverride(node, enclosingType, overridenMethod, overridenType, p);

        }

        // check if return type is primitive readonly
        if (methodElt.getReturnType().getKind().isPrimitive()  // primitive
            && (method.getReturnType().hasAnnotation(READONLY)    // illegal annotation
                || method.getReturnType().hasAnnotation(ROMAYBE)))
            checker.report(Result.failure("primitive.ro"), node);

        return super.visitMethod(node, p);

    }

    /**
     * Overrides its super method, and does not invoke it, since currently BaseTypeVisitor does not
     * have a signature for this method that accepts AnnotatedTypeMirror parameters for both variable
     * and parent.
     *
     * On its current implementation, this method is responsible for assigning implicit annotations
     * to this-mutable variables, depending on receiver type, and to check that fields of ReadOnly
     * objects are not reassigned.
     */
    @Override
    protected void commonAssignmentCheck(Tree varTree, ExpressionTree valueTree, String errorKey, Void p) {

        AnnotatedTypeMirror variable = factory.getAnnotatedType(varTree),
            value = factory.getAnnotatedType(valueTree);

        boolean variableLocalField, valueLocalField;

        // we want to determine if the variable is a field of the local class,
        // so if the current class has a field f,
        // a.f is not a local field, while this.f and f are local fields.
        if (varTree.getKind() == Tree.Kind.MEMBER_SELECT) {
            IdentifierTree varExprTree = (IdentifierTree) ((MemberSelectTree)varTree).getExpression();
            variableLocalField = factory.elementFromUse(varExprTree).getKind().isField();
        } else {
            variableLocalField = variable.getElement().getKind().isField();
        }

        AnnotatedTypeMirror valueType = factory.getAnnotatedType(valueTree);
        if (valueTree.getKind() == Tree.Kind.MEMBER_SELECT) {
            IdentifierTree valueExprTree = (IdentifierTree) ((MemberSelectTree)valueTree).getExpression();
            valueLocalField = factory.elementFromUse(valueExprTree).getKind().isField();
        } else {
            if (valueType.getElement() == null) {
                valueLocalField = false;
            } else {
                valueLocalField = valueType.getElement().getKind().isField();
            }
        }

        // variable must be marked same as state
        AnnotatedDeclaredType mReceiver = visitorState.getMethodReceiver();
        if (mReceiver != null) {
        	if (variableLocalField) {
        		if (!(variable.hasAnnotation(MUTABLE) || variable.hasAnnotation(ROMAYBE) || variable.hasAnnotation(READONLY))) {
        			if (mReceiver.hasAnnotation(READONLY))
        				variable.addAnnotation(READONLY);
        			else if (mReceiver.hasAnnotation(ROMAYBE))
        				variable.addAnnotation(ROMAYBE);
        			else
        				variable.addAnnotation(MUTABLE);
        		}
        	}
        	if (valueLocalField) {
        		if (!(value.hasAnnotation(MUTABLE) || value.hasAnnotation(ROMAYBE) || value.hasAnnotation(READONLY))) {
        			if (mReceiver.hasAnnotation(READONLY))
        				value.addAnnotation(READONLY);
        			else if (mReceiver.hasAnnotation(ROMAYBE))
        				value.addAnnotation(ROMAYBE);
        			else
        				value.addAnnotation(MUTABLE);
        		}
        	}
        }

    	boolean inConstructor = inConstructor();

    	if (mReceiver != null)
    		if ((mReceiver.hasAnnotation(READONLY) || mReceiver.hasAnnotation(ROMAYBE)) && !inConstructor)
    			if (!variable.hasAnnotation(ASSIGNABLE) && variableLocalField)
    				checker.report(Result.failure("ro.field"), varTree);

        if (varTree.getKind() == Tree.Kind.MEMBER_SELECT) {
            ExpressionTree et = ((MemberSelectTree)varTree).getExpression();
            AnnotatedTypeMirror outerType = factory.getAnnotatedType(et);
            Element exprElt = outerType.getElement();

            if (!(variableLocalField && inConstructor)
                && exprElt.getEnclosingElement().getKind()
                != ElementKind.METHOD) {
                if (!variable.hasAnnotation(ASSIGNABLE)
                    && (outerType.hasAnnotation(READONLY)
                        || outerType.hasAnnotation(QREADONLY)
                        || outerType.hasAnnotation(ROMAYBE)))
                    checker.report(Result.failure("ro.field"),varTree);

            }
        }

        // do not check primitives
        if (variable.getKind().isPrimitive())
            return;

        boolean success = checker.isSubtype(variable, value);

        //String useKey = (errorKey == null ? "type.incompatible" : errorKey);
        String useKey = "type.incompatible";
        if (!success) {
            checker.report(Result.failure(useKey,
                    value.toString(), variable.toString()), valueTree);
        }

    }

    /**
     * Checks for prefix and postfix increments and decrements on final fields.
     *
     * @see com.sun.source.util.TreeScanner#visitUnary(com.sun.source.tree.UnaryTree,java.lang.Object)
     */
    @Override
    public Void visitUnary(UnaryTree node, Void p){
        Tree.Kind kind = node.getKind();
        if (visitorState.getMethodReceiver().hasAnnotation(MUTABLE) && !inConstructor()) {
            if (kind == Tree.Kind.POSTFIX_DECREMENT
                || kind == Tree.Kind.POSTFIX_INCREMENT
                || kind == Tree.Kind.PREFIX_DECREMENT
                || kind == Tree.Kind.PREFIX_INCREMENT) {

                AnnotatedTypeMirror variable
                    = factory.getAnnotatedType(node.getExpression());
                if (!variable.hasAnnotation(ASSIGNABLE)
                    && ElementFilter
                    .fieldsIn(visitorState.getClassType().getElement().getEnclosedElements())
                    .contains(variable.getElement())) { // if local field
                    checker.report(Result.failure("ro.field"),
                                   node.getExpression());
                }
            }
        }
        return super.visitUnary(node, p);
    }


    /**
     * Checks primitive variables for illegal annotations.
     *
     * @see com.sun.source.util.TreeScanner#visitUnary(com.sun.source.tree.VariableTree,java.lang.Object)
     */
    @Override
    public Void visitVariable(VariableTree node, Void p){
        AnnotatedTypeMirror v = factory.getAnnotatedType(node);
        if (v.getKind().isPrimitive()
            && (v.hasAnnotation(READONLY)
                || v.hasAnnotation(QREADONLY)
                || v.hasAnnotation(ROMAYBE)))
            checker.report(Result.failure("primitive.ro"), node);

        return super.visitVariable(node, p);
    }

    /**
     * Emits a warning if a cast is made increasing mutability.
     *
     * @see com.sun.source.util.TreeScanner#visitTypeCast(com.sun.source.Tree.TypeCastTree,java.lang.Object)
     */
    @Override
    public Void visitTypeCast(TypeCastTree node, Void p) {

        AnnotatedTypeMirror exprType = factory.getAnnotatedType(node.getExpression()),
            castType = factory.getAnnotatedType(node);

        if (!checker.isSubtype(castType, exprType)) {
            checker.report(Result.warning("mutable.cast"), node.getType());
        }

        return super.visitTypeCast(node, p);
    }

    /**
     * Overriding invocability check to let {code @ReadOnly} references invoke {code @RoMaybe} methods.
     */
    @Override
    protected boolean checkMethodInvocability(AnnotatedExecutableType method, MethodInvocationTree node) {
        AnnotatedTypeMirror methodReceiver = method.getReceiverType().getErased();
        AnnotatedTypeMirror treeReceiver;

        if (methodReceiver.hasAnnotation(ROMAYBE)) methodReceiver.addAnnotation(READONLY);

        if (node.getMethodSelect().getKind() == Tree.Kind.MEMBER_SELECT) {
            MemberSelectTree mst = (MemberSelectTree) (node.getMethodSelect());
            treeReceiver = factory.getAnnotatedType(mst.getExpression());
        } else
            treeReceiver = visitorState.getClassType();


        if (!checker.isSubtype(methodReceiver, treeReceiver)) {
            checker.report(Result.failure("ro.reference",
                factory.elementFromUse(node),
                treeReceiver.toString(), methodReceiver.toString()), node);
            return false;
        }
        return true;
    }


    /**
     * Overriding assignment check to annotate every {code @RoMaybe} parameter with {code @ReadOnly}.
     */
    @Override
    protected void checkArguments(AnnotatedExecutableType method,
            List<? extends ExpressionTree> passedArgs, Void p) {
        List<AnnotatedTypeMirror> requiredArgs = method.getParameterTypes();
        int maxCheck = (method.isVarArgs() ? requiredArgs.size() - 1 : requiredArgs.size());

        for (int i = 0; i < maxCheck; ++i) {
        	AnnotatedTypeMirror pType = requiredArgs.get(i);
        	if (pType.hasAnnotation(ROMAYBE) && !pType.hasAnnotation(MUTABLE))
        		pType.addAnnotation(READONLY);
        	else if (!pType.hasAnnotation(READONLY))
        		pType.addAnnotation(MUTABLE);

        	commonAssignmentCheck(pType,
                    passedArgs.get(i),
                    "argument.invalid", p);
        }

        if (method.isVarArgs()) {
            AnnotatedArrayType vararg = (AnnotatedArrayType) requiredArgs.get(requiredArgs.size() - 1);
            AnnotatedTypeMirror varComponent = vararg.getComponentType();
            if (varComponent.hasAnnotation(ROMAYBE) && !varComponent.hasAnnotation(MUTABLE))
            	varComponent.addAnnotation(READONLY);
            else if (!varComponent.hasAnnotation(READONLY))
            	varComponent.addAnnotation(MUTABLE);

            for (int i = maxCheck; i < passedArgs.size(); ++i) {
                commonAssignmentCheck(varComponent,
                        passedArgs.get(i),
                        "argument.invalid", p);
            }
        }
    }


    private boolean inConstructor() {
    	MethodTree mt = visitorState.getMethodTree();
    	if (mt == null) return true;
    	ExecutableElement mElt = factory.elementFromDeclaration(mt);
    	return mElt.getKind() == ElementKind.CONSTRUCTOR;
    }

}
