package checkers.util;

import static javax.lang.model.util.ElementFilter.methodsIn;

import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Deque;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.annotation.processing.ProcessingEnvironment;
import javax.lang.model.element.*;
import javax.lang.model.type.TypeKind;

import com.sun.source.tree.MethodInvocationTree;
import com.sun.source.tree.Tree;

import checkers.types.*;
import checkers.types.AnnotatedTypeMirror.*;
import checkers.quals.*;

/**
 * Utility methods for operating on {@code AnnotatedTypeMirror}. This
 * class mimics the class {@link javax.lang.model.util.Types}.
 */
public class AnnotatedTypes {

    private ProcessingEnvironment env;
    private AnnotatedTypeFactory factory;

    /**
     * Constructor for {@code AnnotatedTypeUtils}
     * 
     * @param env
     *            the processing environment for this round
     */
    public AnnotatedTypes(ProcessingEnvironment env, AnnotatedTypeFactory factory) {
        this.env = env;
        this.factory = factory;
    }

    /**
     * Returns the most specific base type of {@code t} that start
     * with the given {@code Element}
     * 
     * @param type      a type
     * @param element   an element
     * @return the base type of t of the given element
     */
    public AnnotatedTypeMirror asSuper(AnnotatedTypeMirror type,
            AnnotatedTypeMirror element) {
        return asSuper.visit(type, element);
    }

    private SimpleAnnotatedTypeVisitor<AnnotatedTypeMirror, AnnotatedTypeMirror> asSuper =
        new SimpleAnnotatedTypeVisitor<AnnotatedTypeMirror, AnnotatedTypeMirror>() {
        
        @Override
        protected AnnotatedTypeMirror defaultAction(AnnotatedTypeMirror type, AnnotatedTypeMirror p) {
            return type;
        }
        
        public AnnotatedTypeMirror visitTypeVariable(AnnotatedTypeVariable type, AnnotatedTypeMirror p) {
            // Operate on the upper bound
            return asSuper(((AnnotatedTypeVariable) type).getUpperBound(),
                    p);
        }
        
        public AnnotatedTypeMirror visitArray(AnnotatedArrayType type, AnnotatedTypeMirror p) {
            // Check if array component is subtype of the element
            // first
            if (shouldStop(p, type))
                return type;
            for (AnnotatedTypeMirror st : type.directSuperTypes()) {
                AnnotatedTypeMirror x = asSuper(st, p);
                if (x != null)
                    return x;
            }
            return null;
        }
        
        public AnnotatedTypeMirror visitDeclared(AnnotatedDeclaredType type, AnnotatedTypeMirror p) {
            // If visited Element is the desired one, we are done
            if (shouldStop(p, type))
                return type;

            // Visit the superclass first!
            for (AnnotatedDeclaredType st : type.directSuperTypes()) {
                if (st.getKind() == TypeKind.DECLARED) {
                    AnnotatedDeclaredType x = (AnnotatedDeclaredType) asSuper(st, p);
                    if (x != null)
                        return x;
                }
            }

            return null;
        }
    };
    
    /**
     * Return the base type of t or any of its outer types that starts
     * with the given type. If none exists, return null.
     * 
     * @param t     a type
     * @param elem   a type
     */
    public AnnotatedTypeMirror asOuterSuper(AnnotatedTypeMirror t,
            AnnotatedTypeMirror elem) {
        switch (t.getKind()) {
        case DECLARED:
            do {
                // Search among supers for a desired supertype
                AnnotatedTypeMirror s = asSuper(t, elem);
                if (s != null)
                    return s;
                // if not found immediately, try enclosing type
                // like A in A.B
                t = t.getEnclosingType();
            } while (t != null && t.getKind() == TypeKind.DECLARED);
            return null;
        case ARRAY:     // intentional follow-through
        case TYPEVAR:   // intentional follow-through
        case WILDCARD:
            return asSuper(t, elem);
        default:
            return null;
        }
    }


    private boolean shouldStop(AnnotatedTypeMirror sup, AnnotatedTypeMirror sub) {
        // Check if it's the same type
        // if sup is primitive, but not sub
        if (sup.getKind().isPrimitive() && !sub.getKind().isPrimitive())
            return true;
        // if both are declared
        if (sup.getKind() == TypeKind.DECLARED && sub.getKind() == TypeKind.DECLARED) {
            AnnotatedDeclaredType supdt = (AnnotatedDeclaredType) sup;
            AnnotatedDeclaredType subdt = (AnnotatedDeclaredType) sub;
            
            // Check if it's the same name
            if (!supdt.getUnderlyingType().asElement().equals(
                    subdt.getUnderlyingType().asElement()))
                return false;
            
            return supdt.getTypeArguments().size() >= subdt.getTypeArguments().size();
        }
        
        if (sup.getKind() == TypeKind.ARRAY && sub.getKind() == TypeKind.ARRAY) {
            AnnotatedArrayType supat = (AnnotatedArrayType) sup;
            AnnotatedArrayType subat = (AnnotatedArrayType) sub;
            return shouldStop(supat.getComponentType(), subat.getComponentType());
        }
        
        return false;
    }
    /**
     * A Place-holder for a method that returns the appropriate
     * {@code AnnotatedType} for the given element
     * 
     * @param elem
     *            an element
     * @return the type associated with the element
     */
    public AnnotatedTypeMirror getAnnotatedType(Element elem) {
        return factory.getAnnotatedType(elem);
    }

    /**
     * The type of given element, seen as a member of t.
     * 
     * @param t    a type
     * @param elem  an element
     */
    public AnnotatedTypeMirror asMemberOf(AnnotatedTypeMirror t, Element elem) {
        if (ElementUtils.isStatic(elem))
            return getAnnotatedType(elem);

        // For Type Variable, operate on the upper
        if (t.getKind() == TypeKind.TYPEVAR && 
                ((AnnotatedTypeVariable)t).getUpperBound() != null)
            return asMemberOf(((AnnotatedTypeVariable) t).getUpperBound(),
                    elem);

        // I cannot think of why it wouldn't be a declared type!
        // Defensive Programming
        if (t.getKind() != TypeKind.DECLARED) {
            return getAnnotatedType(elem);
        }

        //
        // Basic Algorithm:
        // 1. Find the owner of the element
        // 2. Find the base type of owner (e.g. type of owner as supertype 
        //      of passed type)
        // 3. Subsitute for type variables if any exist
        TypeElement owner = ElementUtils.enclosingClass(elem);
        AnnotatedDeclaredType ownerType =
            (AnnotatedDeclaredType) getAnnotatedType(owner);

        // TODO: Potential bug if Raw type is used
        if (ElementUtils.isStatic(elem) || owner.getTypeParameters().isEmpty())
            return getAnnotatedType(elem);

        AnnotatedDeclaredType base =
            (AnnotatedDeclaredType) asOuterSuper(t, ownerType);

        if (base == null)
            return getAnnotatedType(elem);

        List<? extends AnnotatedTypeMirror> ownerParams = 
            ownerType.getTypeArguments();
        List<? extends AnnotatedTypeMirror> baseParams =
            base.getTypeArguments();
        if (!ownerParams.isEmpty()) {
            if (baseParams.isEmpty())
                return getAnnotatedType(elem).getErased();
            else
                return subst(getAnnotatedType(elem), ownerParams, baseParams);
        }

        return getAnnotatedType(elem);

    }

    /**
     * Returns a new type, a copy of the passed {@code t}, with all
     * instances of {@code from} type substituted with their correspondents
     * in {@code to} and return the substituted in type.
     * 
     * @param t     the type
     * @param from  the from types
     * @param to    the to types
     * @return  the new type after substitutions
     */
    public AnnotatedTypeMirror subst(AnnotatedTypeMirror t,
            List<? extends AnnotatedTypeMirror> from,
            List<? extends AnnotatedTypeMirror> to) {
        assert from.size() == to.size();
        Map<AnnotatedTypeMirror, AnnotatedTypeMirror> mappings =
            new HashMap<AnnotatedTypeMirror, AnnotatedTypeMirror>();

        for (int i = 0; i < from.size(); ++i) {
            mappings.put(from.get(i), to.get(i));
        }
        return t.substitute(mappings);
    }


    /**
     * Returns a new type, a copy of the passed {@code t}, with all
     * instances of {@code mappings.keySet()} type substituted with 
     * their correspondents in {@code to} and return the substituted 
     * in type.
     * 
     * @param ts        the types
     * @param mappings  the mapping for the types
     * @return  the new type after substitutions
     */
    public List<? extends AnnotatedTypeMirror> subst(
            List<? extends AnnotatedTypeMirror> ts,
            Map<? extends AnnotatedTypeMirror, ? extends AnnotatedTypeMirror> mappings) {
        List<AnnotatedTypeMirror> rTypes = new ArrayList<AnnotatedTypeMirror>();
        for (AnnotatedTypeMirror type : ts) 
            rTypes.add(type.substitute(mappings));
        return rTypes;
    }

    /**
     * Returns a deep copy of the passed type
     * 
     * @param type  the annotated type to be copied
     * @return a deep copy of the passed type
     */
    public AnnotatedTypeMirror deepCopy(AnnotatedTypeMirror type) {
        // TODO: Test this, specify behaviour
        return type.substitute(Collections.<AnnotatedTypeMirror, 
                AnnotatedTypeMirror>emptyMap());
    }
    
    /**
     * Returns the iterated type of the passed type: either the component type
     * of an array if {@code iterable} is an array, or the type argument
     */
    public AnnotatedTypeMirror getIteratedType(AnnotatedTypeMirror iterableType) {
        if (iterableType.getKind() == TypeKind.ARRAY) {
            return ((AnnotatedArrayType) iterableType).getComponentType();
        } else if (iterableType.getKind() == TypeKind.DECLARED) {
            Element iterableElement = env.getElementUtils().getTypeElement("java.lang.Iterable");
            AnnotatedDeclaredType iterableElmType = (AnnotatedDeclaredType) getAnnotatedType(iterableElement);
            AnnotatedDeclaredType dt = (AnnotatedDeclaredType) asSuper(iterableType, iterableElmType);
            if (dt == null)
                return null;
            else if (dt.getTypeArguments().isEmpty()) {
                // was erased
                return this.getAnnotatedType(env.getElementUtils().getTypeElement("java.lang.Object"));
            } else {
                return dt.getTypeArguments().get(0);
            }
        } else {
            // Type is not iterable
            return null;
        }
    }
    
    public List<AnnotatedDeclaredType> getSuperTypes(AnnotatedDeclaredType subtype) {
        
        List<AnnotatedDeclaredType> supertypes = new ArrayList<AnnotatedDeclaredType>();
        if (subtype == null) 
            return supertypes;

        // Set up a stack containing the type mirror of subtype, which
        // is our starting point.
        Deque<AnnotatedDeclaredType> stack = new ArrayDeque<AnnotatedDeclaredType>();
        stack.push(subtype);

        while (!stack.isEmpty()) {
            AnnotatedDeclaredType current = stack.pop();

            // For each direct supertype of the current type, if it
            // hasn't already been visited, push it onto the stack and
            // add it to our supertypes set.
            for (AnnotatedDeclaredType supertype : current.directSuperTypes()) {
                if (!supertypes.contains(supertype)) {
                    stack.push(supertype);
                    supertypes.add(supertype);
                }
            }
        }

        return Collections.<AnnotatedDeclaredType>unmodifiableList(supertypes);
    }

    /**
     * A utility method that takes a Method element and returns a set
     * of all elements that this method overrides (as
     * {@link javax.lang.model.element.ExecutableElement}s)
     * 
     * @param method
     *            the overriding method
     * @return an unmodifiable set of {@link ExecutableElement}s
     *         representing the elements that method overrides
     */
    public Map<AnnotatedDeclaredType, ExecutableElement> overriddenMethods(
            ExecutableElement method) {
        final TypeElement elem = (TypeElement) method.getEnclosingElement();
        final AnnotatedDeclaredType type = 
            (AnnotatedDeclaredType) getAnnotatedType(elem);
        final List<AnnotatedDeclaredType> supertypes = getSuperTypes(type);
        return overriddenMethods(method, supertypes);
    }

    /**
     * A utility method that takes the element for a method and the
     * set of all supertypes of the method's containing class and
     * returns the set of all elements that method overrides (as
     * {@link javax.lang.model.element.ExecutableElement}s).
     * 
     * @param method
     *            the overriding method
     * @param supertypes
     *            the set of supertypes to check for methods that are
     *            overriden by {@code method}
     * @return an unmodified set of {@link ExecutableElement}s
     *         representing the elements that {@code method} overrides
     *         among {@code supertypes}
     */
    public Map<AnnotatedDeclaredType, ExecutableElement> overriddenMethods(
            ExecutableElement method, List<AnnotatedDeclaredType> supertypes) {

        Map<AnnotatedDeclaredType, ExecutableElement> overrides =
            new HashMap<AnnotatedDeclaredType, ExecutableElement>();
        
        for (AnnotatedDeclaredType supertype : supertypes) {
            /*@Nullable*/ TypeElement superElement = 
                (TypeElement) supertype.getUnderlyingType().asElement();
            assert superElement != null; /*nninvariant*/
            // For all method in the supertype, add it to the set if
            // it overrides the given method.
            for (ExecutableElement supermethod : methodsIn(superElement
                    .getEnclosedElements())) {
                if (env.getElementUtils().overrides(method, supermethod,
                        superElement)) {
                    overrides.put(supertype, supermethod);
                    break;
                }
            }
        }

        return Collections.</*@NonNull*/ AnnotatedDeclaredType,
            /*@NonNull*/ ExecutableElement>unmodifiableMap(overrides);
    }

    public Map<TypeParameterElement, AnnotatedTypeMirror> 
    findTypeParameters(MethodInvocationTree methodInvocation) {
        Map<TypeParameterElement, AnnotatedTypeMirror> mapping = 
            new HashMap<TypeParameterElement, AnnotatedTypeMirror>();
        ExecutableElement methodElt = factory.elementFromUse(methodInvocation);
        if (methodElt.getTypeParameters().isEmpty())
            return mapping;
        if (!methodInvocation.getTypeArguments().isEmpty()) {
            // User already supplied this! yay!
            for (int i = 0; i < methodElt.getTypeParameters().size(); ++i) {
                TypeParameterElement el = methodElt.getTypeParameters().get(i);
                Tree tree = methodInvocation.getTypeArguments().get(i);
                AnnotatedTypeMirror type = factory.fromTypeTree(tree);
                mapping.put(el, type);
            }
        }
        return mapping;
    }
}
