package tests;

import javax.tools.*;
import java.io.*;
import java.util.*;

import static org.junit.Assert.*;

import checkers.quals.*;

/**
 * Abstract class for testing a checker framework.
 */
abstract public class CheckerTest {

    /** The fully-qualified class name of the checker to use for tests. */
    protected final String checkerName;

    /** The path to the directory containing test inputs. */
    protected final String checkerDir;

    /** Extra options to pass to javac when running the checker. */
    protected final String[] checkerOptions;

    /**
     * Creates a new checker test.
     *
     * @param checkerName the fully-qualified class name of the checker to use
     * @param checkerDir the path to the directory of test inputs
     * @param checkerOptions options to pass to the compiler when running tests
     */
    public CheckerTest(String checkerName, String checkerDir, String... checkerOptions) {
        this.checkerName = checkerName;
        this.checkerDir = checkerDir;
        this.checkerOptions = checkerOptions;
    }

    /**
     * Runs a test, using the {@code shouldPass} parameter to determine
     * whether or not the test should pass. The method uses reflection to
     * determine the expected output and Java source files that would
     * otherwise be passed to {@link #runTest} -- if the calling method is
     * named testZZZ, this method uses an expected outfile called "ZZZ.out"
     * and an Java source file called "ZZZ.java".
     *
     * @param shouldPass whether or not the test should pass
     */
    protected void test(boolean shouldPass) {
        StackTraceElement[] stack = Thread.currentThread().getStackTrace();
        assert stack.length >= 3;
        String method = stack[2].getMethodName();
        if (!method.startsWith("test"))
            throw new AssertionError("caller's name is invalid");
        String[] parts = method.split("test");
        String test = parts[parts.length - 1];
        runTest(test + ".out", shouldPass, test + ".java");
    }

    /**
     * Compiles and returns a TestRun.
     */
    protected TestRun getTest(String... files) {

        List<String> fileStrings = new LinkedList<String>();
        for (String s : files)
            fileStrings.add(checkerDir + "/" + s);

        JavaCompiler compiler = ToolProvider.getSystemJavaCompiler();
        StandardJavaFileManager fileManager
            = compiler.getStandardFileManager(null, null, null);
        Iterable<? extends JavaFileObject> tests
            = fileManager.getJavaFileObjectsFromStrings(fileStrings);

        return
            TestInput.compileAndCheck(tests, checkerName, checkerOptions);
    }

    /**
     * Compiles and returns a TestRun.
     */
    protected TestRun getTest(File... files) {
        JavaCompiler compiler = ToolProvider.getSystemJavaCompiler();
        StandardJavaFileManager fileManager
            = compiler.getStandardFileManager(null, null, null);
        Iterable<? extends JavaFileObject> tests
            = fileManager.getJavaFileObjects(files);

        return TestInput.compileAndCheck(tests, checkerName, checkerOptions);
    }

    /**
     * Tests that the result of compiling the javaFile matches the expectedFile.
     *
     * @param expectedFile  the expected result for compilation
     * @param shouldSucceed whether the javaFile should compile successfully
     * @param javaFiles  the Java files to be compiled
     */
    protected void runTest(String expectedFile, boolean shouldSucceed, String ... javaFiles) {
        TestRun run = getTest(javaFiles);
        checkTestResult(run, expectedFile, shouldSucceed);
    }

    /**
     * Tests that the result of compiling the javaFile matches the expectedFile.
     *
     * @param expectedFile  the expected result for compilation
     * @param shouldSucceed whether the javaFile should compile successfully
     * @param javaFiles  the Java files to be compiled
     */
     protected void runTest(String expectedFile, boolean shouldSucceed, File ... javaFiles) {
        TestRun run = getTest(javaFiles);
        checkTestResult(run, expectedFile, shouldSucceed);
    }

    protected void checkTestResult(TestRun run, String expectedFileName, boolean shouldSucceed) {
        assert expectedFileName != null;
        if (shouldSucceed)
            assertSuccess(run);
        else
            assertFailure(run);
        File expectedFile = new File(checkerDir + File.separator + expectedFileName);
        if ((!shouldSucceed) && !(expectedFile.exists())) {
            throw new Error("Did not find expected file " + expectedFile);
        }
        if (shouldSucceed && !(expectedFile.exists())) {
            return;
        }
        List<Diagnostic<? extends JavaFileObject>> list = run.getDiagnostics();
        assertDiagnostics(list, expectedFileName);
    }


    /**
     * Asserts that the test compilation completed without failures or
     * exceptions.
     *
     * @param run the test run to check
     */
    protected void assertSuccess(/*@ReadOnly*/ TestRun run) {
        assertTrue("failures:\n" + run.getDiagnostics(), run.getResult());
    }

    /**
     * Asserts that the test compilation did not complete successfully.
     *
     * @param run the test run to check
     */
    protected void assertFailure(/*@ReadOnly*/ TestRun run) {
        assertFalse("the test run was successful", run.getResult());
    }

    /**
     * Compares the result of the compiler against a list of errors in a file.
     * If the file is not found or cannot be read, the assertion fails.
     *
     * @param list the list of diagnostics from the compiler
     * @param diagnosticFile a file containing a list of expected errors, one
     *        per line
     */
    protected void assertDiagnostics(/*@ReadOnly*/ List</*@ReadOnly*/ Diagnostic<? extends JavaFileObject>> list,
            /*@ReadOnly*/ String diagnosticFile) {

        try {
            String filePath = checkerDir + File.separator + diagnosticFile;
            BufferedReader reader = new BufferedReader(new FileReader(filePath));
            ArrayList<String> lines = new ArrayList<String>();
            String line;
            while ((line = reader.readLine()) != null) {
                if (line.startsWith("#"))
                    continue;
                lines.add(line.substring(line.indexOf(':')).trim());
            }
            assertDiagnostics(list, lines.toArray(new String[lines.size()]));
        } catch (IOException e) {
            fail(e.getMessage());
        }
    }

    /**
     * Compares the result of the compiler against an array of Strings.
     */
    protected void assertDiagnostics(/*@ReadOnly*/ List</*@ReadOnly*/ Diagnostic<? extends JavaFileObject>> list, /*@ReadOnly*/ String[/*@ReadOnly*/] diagnostics) {
        String cs = (checkerDir == "" ? "" : checkerDir + "/"); // "interned"

        List<String> expectedList = new LinkedList<String>();
        for (/*@ReadOnly*/ String sd : diagnostics) expectedList.add(/* cs + */ sd);

        List<String> resultsList = new LinkedList<String>();
        for (/*@ReadOnly*/ Diagnostic<? extends JavaFileObject> d : list) {
            String result = d.toString().trim();
            if (result.contains("\n")){
                result = result.substring(0, result.indexOf('\n'));
            }
            result = result.substring(result.indexOf(':')).trim();
            resultsList.add(result);
        }

        List<String> foundList = new LinkedList<String>();
        foundList.addAll(resultsList);
        foundList.retainAll(expectedList);

        String failMessage = foundList.size() + " out of "
            + expectedList.size() + " expected diagnostics "
            + (foundList.size() == 1 ? "was" : "were") +" found.\n";
        boolean failed = false;

        List<String> notFoundList = new LinkedList<String>();
        notFoundList.addAll(expectedList);
        notFoundList.removeAll(resultsList);

        if (!notFoundList.isEmpty()) {
            failed = true;

            String message = notFoundList.size() == 1?
                "1 expected diagnostic was not found:\n" :
                notFoundList.size() + " expected diagnostics were not found:\n";
            failMessage += "\n" + message;

            for (String a : notFoundList)
                failMessage += a + "\n";

        }


        List<String> unexpectedList = new LinkedList<String>();
        unexpectedList.addAll(resultsList);
        unexpectedList.removeAll(expectedList);

        if (!unexpectedList.isEmpty()) {
            failed = true;

            String message = unexpectedList.size() == 1?
                "1 unexpected diagnostic was found:\n" :
                unexpectedList.size() + " unexpected diagnostics were found:\n";
            failMessage += "\n" + message;

            for (String a : unexpectedList)
                failMessage += a + "\n";
        }

        if (failed)
            fail(failMessage);

    }

}
